import datetime
import typing as t

import pytest

from izulu import _utils
from tests import errors
from tests import helpers as h

count = 42
owner = "somebody"
dt = datetime.datetime.now(datetime.timezone.utc)


@pytest.mark.parametrize(
    ("kwargs", "expected"),
    [
        (h._make_store_kwargs(), frozenset()),
        (
            h._make_store_kwargs(fields=("field_1", "field_2")),
            frozenset(("field_1", "field_2")),
        ),
        (
            h._make_store_kwargs(inst_hints=dict(field_2=2, field_3=3)),
            frozenset(("field_2", "field_3")),
        ),
        (
            h._make_store_kwargs(
                fields=("field_1", "field_2"),
                inst_hints=dict(field_2=2, field_3=3),
            ),
            frozenset(("field_1", "field_2", "field_3")),
        ),
        (
            h._make_store_kwargs(
                fields=("field_1", "field_2"),
                inst_hints=dict(field_1=1, field_2=2),
            ),
            frozenset(("field_1", "field_2")),
        ),
    ],
)
def test_store_post_init(kwargs, expected):
    assert _utils.Store(**kwargs).registered == expected


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        ((tuple(),), ""),
        ((("item_1",),), "'item_1'"),
        ((("item_1", "item_2", "item_3"),), "'item_1', 'item_2', 'item_3'"),
    ],
)
def test_join_items(args, expected):
    assert _utils.join_items(*args) == expected


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        (dict(), ""),
        (dict(a=42), "a=42"),
        (
            dict(owner=owner, count=count, timestamp=dt),
            f"{owner=!r}, {count=!r}, timestamp={dt!r}",
        ),
        (dict(timestamp=dt), f"timestamp={dt!r}"),
    ],
)
def test_join_kwargs(data, expected):
    assert _utils.join_kwargs(**data) == expected


@pytest.mark.parametrize(
    ("template", "kwargs", "expected"),
    [
        ("Static message template", dict(), "Static message template"),
        (
            "Static message template",
            dict(name="John", age=42, ENTITY="The thing"),
            "Static message template",
        ),
        (
            "The {name} and {ENTITY}",
            dict(name="John", age=42, ENTITY="The thing"),
            "The John and The thing",
        ),
        (
            "The {name} of {age} and {ENTITY}",
            dict(name="John", age=42, ENTITY="The thing"),
            "The John of 42 and The thing",
        ),
        (
            "The {name} of {age:f} and {ENTITY}",
            dict(name="John", age=42, ENTITY="The thing"),
            "The John of 42.000000 and The thing",
        ),
    ],
)
def test_format_template_ok(template, kwargs, expected):
    assert _utils.format_template(template, kwargs) == expected


@pytest.mark.parametrize(
    ("template", "kwargs"),
    [
        (
            "The {name} of {age} and {ENTITY}",
            dict(name="John", ENTITY="The thing"),
        ),
        (
            "The {name} of {age:f} and {ENTITY}",
            dict(name="John", age="42", ENTITY="The thing"),
        ),
    ],
)
def test_format_template_fail(template, kwargs):
    with pytest.raises(
        ValueError, match="Failed to format template with provided kwargs:"
    ):
        _utils.format_template(template, kwargs)


@pytest.mark.parametrize(
    ("tpl", "expected"),
    [
        ("", tuple()),
        ("abc def", tuple()),
        ("{}", ("",)),
        ("abc {} def", ("",)),
        ("{-}", ("-",)),
        ("{_}", ("_",)),
        ("{'}", ("'",)),
        ("{01abc}", ("01abc",)),
        ("{]}", ("]",)),
        ("{1}", (1,)),
        ("{field}", ("field",)),
        ("{field.attribute}", ("field",)),
        ("{field[key]}", ("field",)),
        ("{field[key].attr}", ("field",)),
        ("{field.attr[key]}", ("field",)),
        ("{field.attr[key].attr}", ("field",)),
        ("{field[key][another]}", ("field",)),
        ("{field.attribute.attr}", ("field",)),
        ("{fi-eld}", ("fi-eld",)),
        ("{field.-attribute}", ("field",)),
        ("{field[-key]}", ("field",)),
        ("{-field[key].attr}", ("-field",)),
        ("{field-.-attr[-key]}", ("field-",)),
        ("{fi-eld.at-tr[key].attr}", ("fi-eld",)),
        ("{field[0key][another]}", ("field",)),
        ("{fi-eld.0attribute.attr}", ("fi-eld",)),
        ("Having boring message here", tuple()),
        ("Hello {}!", ("",)),
        ("Hello {0}!", (0,)),
        ("Hello {you}!", ("you",)),
        ("Hello {you:f}!", ("you",)),
        ("Hello {you}! How are you, {you!a}", ("you", "you")),
        (
            "{owner:f!a}: Having {!a} count={count!a:f} for owner={0:f}",
            ("owner", "", "count", 0),
        ),
    ],
)
def test_iterate_field_specs(tpl, expected):
    assert tuple(_utils.iter_fields(tpl)) == expected


@pytest.mark.parametrize(
    ("kls", "const_hints", "inst_hints"),
    [
        (errors.RootError, dict(), dict()),
        (errors.TemplateOnlyError, dict(), dict()),
        (errors.ComplexTemplateOnlyError, dict(), dict()),
        (errors.AttributesOnlyError, dict(), dict(age=int, name=str)),
        (
            errors.AttributesWithStaticDefaultsError,
            dict(),
            dict(age=int, name=str),
        ),
        (
            errors.AttributesWithDynamicDefaultsError,
            dict(),
            dict(age=int, name=str),
        ),
        (
            errors.ClassVarsError,
            dict(
                name=t.ClassVar[str],
                age=t.ClassVar[int],
                blah=t.ClassVar[float],
            ),
            dict(),
        ),
        (
            errors.MixedError,
            dict(entity=t.ClassVar[str]),
            dict(name=str, age=int, my_type=str, timestamp=datetime.datetime),
        ),
        (
            errors.DerivedError,
            dict(entity=t.ClassVar[str]),
            dict(
                name=str,
                surname=str,
                full_name=str,
                age=int,
                my_type=str,
                timestamp=datetime.datetime,
                updated_at=datetime.datetime,
                box=dict,
                location=t.Tuple[float, float],
            ),
        ),
    ],
)
def test_split_cls_hints(kls, const_hints, inst_hints):
    assert _utils.split_cls_hints(kls) == (const_hints, inst_hints)


@pytest.mark.parametrize(
    ("kls", "attrs", "expected"),
    [
        (errors.DerivedError, (), dict()),
        (errors.DerivedError, ("entity",), dict(entity="The Entity")),
        (
            errors.DerivedError,
            (
                "name",
                "surname",
                "age",
                "timestamp",
                "my_type",
                "location",
                "updated_at",
                "full_name",
                "box",
            ),
            dict(
                age=0,
                location=(50.3, 3.608),
                timestamp=errors.DerivedError.timestamp,
                updated_at=errors.DerivedError.updated_at,
                full_name=errors.DerivedError.full_name,
                my_type=errors.DerivedError.my_type,
            ),
        ),
        (errors.RootError, ("entity",), dict()),
        (errors.TemplateOnlyError, ("name",), dict()),
    ],
)
def test_get_cls_defaults(kls, attrs, expected):
    assert _utils.get_cls_defaults(kls, attrs) == expected



import datetime as dtm
import typing as t

from izulu import root


class RootError(root.Error):
    __toggles__ = root.Toggles.DEFAULT ^ root.Toggles.FORBID_UNANNOTATED_FIELDS


class TemplateOnlyError(RootError):
    __template__ = "The {name} is {age} years old"


class ComplexTemplateOnlyError(RootError):
    __template__ = "{name:*^20} {age: f} {age:#b} {ts:%Y-%m-%d %H:%M:%S}"
    __toggles__ = root.Toggles.NONE


class AttributesOnlyError(RootError):
    __template__ = "Static message template"

    name: str
    age: int


class AttributesWithStaticDefaultsError(RootError):
    __template__ = "Static message template"

    name: str
    age: int = 0


class AttributesWithDynamicDefaultsError(RootError):
    __template__ = "Static message template"

    name: str
    age: int = root.factory(default_factory=int)


class ClassVarsError(RootError):
    __template__ = "Static message template"

    name: t.ClassVar[str] = "Username"
    age: t.ClassVar[int] = 42
    blah: t.ClassVar[float]


class MixedError(RootError):
    __template__ = "The {name} is {age} years old with {note}"

    entity: t.ClassVar[str] = "The Entity"

    name: str
    age: int = 0
    timestamp: dtm.datetime = root.factory(default_factory=dtm.datetime.now)
    my_type: str = root.factory(
        default_factory=lambda self: self.__class__.__name__,
        self=True,
    )


class DerivedError(MixedError):
    __template__ = "The {name} {surname} is {age} years old with {note}"

    entity: t.ClassVar[str] = "The Entity"

    surname: str
    location: t.Tuple[float, float] = (50.3, 3.608)
    updated_at: dtm.datetime = root.factory(default_factory=dtm.datetime.now)
    full_name: str = root.factory(
        default_factory=lambda self: f"{self.name} {self.surname}",
        self=True,
    )
    box: dict


class MyError(RootError):
    __template__ = "The {name} is {age} years old with {ENTITY} {note}"

    DEFAULT = "default"
    HINT: t.ClassVar[int]
    ENTITY: t.ClassVar[str] = "The Entity"

    name: str
    age: int = 0
    timestamp: dtm.datetime = root.factory(default_factory=dtm.datetime.now)
    my_type: str = root.factory(
        default_factory=lambda self: self.__class__.__name__,
        self=True,
    )



import datetime
import types
import typing as t
from unittest import mock

import pytest

from izulu import root
from tests import errors


@pytest.mark.parametrize(
    ("kls", "fields", "hints", "registered", "defaults", "consts"),
    [
        (
            errors.RootError,
            frozenset(),
            types.MappingProxyType({}),
            frozenset(),
            frozenset(),
            types.MappingProxyType({}),
        ),
        (
            errors.TemplateOnlyError,
            frozenset(("name", "age")),
            types.MappingProxyType({}),
            frozenset(("name", "age")),
            frozenset(),
            types.MappingProxyType({}),
        ),
        (
            errors.AttributesOnlyError,
            frozenset(),
            types.MappingProxyType(dict(name=str, age=int)),
            frozenset(("name", "age")),
            frozenset(),
            types.MappingProxyType({}),
        ),
        (
            errors.AttributesWithStaticDefaultsError,
            frozenset(),
            types.MappingProxyType(dict(name=str, age=int)),
            frozenset(("name", "age")),
            frozenset(("age",)),
            types.MappingProxyType({}),
        ),
        (
            errors.AttributesWithDynamicDefaultsError,
            frozenset(),
            types.MappingProxyType(dict(name=str, age=int)),
            frozenset(("name", "age")),
            frozenset(("age",)),
            types.MappingProxyType({}),
        ),
        (
            errors.ClassVarsError,
            frozenset(),
            types.MappingProxyType({}),
            frozenset(),
            frozenset(),
            types.MappingProxyType(dict(name="Username", age=42)),
        ),
        (
            errors.MixedError,
            frozenset(("name", "age", "note")),
            types.MappingProxyType(
                dict(
                    name=str, age=int, timestamp=datetime.datetime, my_type=str
                )
            ),
            frozenset(("name", "age", "note", "timestamp", "my_type")),
            frozenset(("age", "timestamp", "my_type")),
            types.MappingProxyType(dict(entity="The Entity")),
        ),
        (
            errors.DerivedError,
            frozenset(("name", "surname", "age", "note")),
            types.MappingProxyType(
                dict(
                    name=str,
                    age=int,
                    timestamp=datetime.datetime,
                    my_type=str,
                    surname=str,
                    location=t.Tuple[float, float],
                    updated_at=datetime.datetime,
                    full_name=str,
                    box=dict,
                )
            ),
            frozenset(
                (
                    "name",
                    "age",
                    "note",
                    "timestamp",
                    "my_type",
                    "surname",
                    "location",
                    "updated_at",
                    "full_name",
                    "box",
                )
            ),
            frozenset(
                (
                    "age",
                    "timestamp",
                    "my_type",
                    "location",
                    "updated_at",
                    "full_name",
                )
            ),
            types.MappingProxyType(dict(entity="The Entity")),
        ),
    ],
)
def test_cls_store(kls, fields, hints, registered, defaults, consts):
    """Validates store management from root.Error.__init_subclass__."""
    store = kls._Error__cls_store

    assert type(store.fields) is type(fields)
    assert store.fields == fields
    assert type(store.inst_hints) is type(hints)
    assert store.inst_hints == hints
    assert type(store.registered) is type(registered)
    assert store.registered == registered
    assert type(store.defaults) is type(defaults)
    assert store.defaults == defaults
    assert type(store.consts) is type(consts)
    assert store.consts == consts


@pytest.mark.parametrize(
    "toggles",
    [root.Toggles.FORBID_NON_NAMED_FIELDS, root.Toggles.NONE],
)
@mock.patch("izulu._utils.check_non_named_fields", return_value=0)
def test_cls_validation(mocked_check, toggles):
    """Validates feature checks from root.Error.__init_subclass__."""
    type("Err", (errors.ClassVarsError,), {"__toggles__": toggles})

    if toggles is root.Toggles.NONE:
        mocked_check.assert_not_called()
    else:
        mocked_check.assert_called_once()



import datetime
from unittest import mock

import pytest

from tests import errors

TS = datetime.datetime.now(datetime.timezone.utc)


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),
        dict(name="John"),
        dict(age=42),
        dict(name="John", age=42),
        dict(name="John", age=42, ts=53452345.3465),
        dict(name="John", age="Karl", ts=TS),
    ],
)
def test_templating(kwargs):
    match = "Failed to format template with provided kwargs:"

    with pytest.raises(ValueError, match=match):
        errors.ComplexTemplateOnlyError(**kwargs)


@pytest.mark.parametrize(
    ("err", "expected"),
    [
        (errors.RootError(), "Unspecified error"),
        (
            errors.MixedError(name="John", age=10, note="..."),
            "The John is 10 years old with ...",
        ),
    ],
)
def test_str(err, expected):
    assert str(err) == expected


@pytest.mark.parametrize(
    ("err", "expected"),
    [
        (errors.RootError(), "tests.errors.RootError()"),
        (
            errors.MixedError(name="John", age=10, note="...", timestamp=TS),
            (
                "tests.errors.MixedError(name='John', age=10,"
                f" note='...', timestamp={TS!r}, my_type='MixedError')"
            ),
        ),
    ],
)
def test_repr(err, expected):
    assert repr(err) == expected


def test_repl_repr():
    e = errors.TemplateOnlyError(name="John", age=42)

    with mock.patch.object(
        errors.TemplateOnlyError,
        "__module__",
        new_callable=mock.PropertyMock,
    ) as mocked:
        mocked.return_value = "__main__"

        assert repr(e) == "__main__.TemplateOnlyError(name='John', age=42)"


@pytest.mark.parametrize(
    ("err", "expected"),
    [
        (errors.RootError(), "RootError: Unspecified error"),
        (
            errors.MixedError(name="John", age=10, note="..."),
            "MixedError: The John is 10 years old with ...",
        ),
    ],
)
def test_as_str(err, expected):
    assert err.as_str() == expected



import types

from izulu import _utils


def _make_store_kwargs(
    fields=None,
    const_hints=None,
    inst_hints=None,
    consts=None,
    defaults=None,
) -> dict:
    return dict(
        fields=frozenset(fields or tuple()),
        const_hints=types.MappingProxyType(const_hints or dict()),
        inst_hints=types.MappingProxyType(inst_hints or dict()),
        consts=types.MappingProxyType(consts or dict()),
        defaults=frozenset(defaults or tuple()),
    )


def _make_store(
    fields=None,
    const_hints=None,
    inst_hints=None,
    consts=None,
    defaults=None,
) -> _utils.Store:
    return _utils.Store(
        **_make_store_kwargs(
            fields=fields,
            inst_hints=inst_hints,
            const_hints=const_hints,
            consts=consts,
            defaults=defaults,
        )
    )
